/**
 * @file
 * @brief Strawberry Linux USBxExvW[ELbg([J[iԁFUSBRH)p̃j^Oc[łB
 * @author AKIYAMA Kouhei
 * @since 2010-07-11
 */
#include <winsock2.h>
#include <windows.h>
#include <tchar.h>
#include <vector>
#include <boost/shared_ptr.hpp>
#include <ctime>
#include <cstdio>
#include <string>

#include "resource.h"

/**
 * USBMeter.dllւ̃ANZX񋟂֐܂Ƃ߂OԂłB
 * ԍŏ usbmeter::LoadUSBMeterDLL ĂяoĂB
 */
namespace usbmeter {
	void ERROR_WINAPI(TCHAR *msg)
	{
		MessageBox(NULL, msg, _T("Error"), MB_OK);
	}

	typedef BSTR (__stdcall *FPGetVers)(BSTR dev);
	typedef BSTR (__stdcall *FPFindUSB)(LONG *index);
	typedef LONG (__stdcall *FPGetTempHumid)(BSTR dev, double *temp, double *humid);
	typedef LONG (__stdcall *FPControlIO)(BSTR dev, LONG port, LONG val);
	typedef LONG (__stdcall *FPSetHeater)(BSTR dev, LONG val);
	typedef LONG (__stdcall *FPGetTempHumidTrue)(BSTR dev, double *temp, double *humid);

	static FPGetVers fpGetVers;
	static FPFindUSB fpFindUSB;
	static FPGetTempHumid fpGetTempHumid;
	static FPControlIO fpControlIO;
	static FPSetHeater fpSetHeater;
	static FPGetTempHumidTrue fpGetTempHumidTrue;

	BSTR GetVers(BSTR dev) { return (*fpGetVers)(dev);}
	BSTR FindUSB(LONG *index) { return (*fpFindUSB)(index);}
	LONG GetTempHumid(BSTR dev, double *temp, double *humid) { return (*fpGetTempHumid)(dev, temp, humid);}
	LONG ControlIO(BSTR dev, LONG port, LONG val) { return (*fpControlIO)(dev, port, val);}
	LONG SetHeater(BSTR dev, LONG val) { return (*fpSetHeater)(dev, val);}
	LONG GetTempHumidTrue(BSTR dev, double *temp, double *humid) { return (*fpGetTempHumidTrue)(dev, temp, humid);}

	template<typename FP>
	bool GetFunction(HMODULE hModule, LPCSTR fname, FP *fpp)
	{
		FP fp = reinterpret_cast<FP>(GetProcAddress(hModule, fname));
		if(!fp){
			ERROR_WINAPI(_T("GetProcAddress Failed."));
			return false;
		}
		*fpp = fp;
		return true;
	}

	bool LoadUSBMeterDLL(void)
	{
		HMODULE hModule = LoadLibrary(_T("USBMeter.dll"));
		if(!hModule){
			ERROR_WINAPI(_T("Failed to load USBMeter.dll. You can download USBMeter.dll from http://strawberry-linux.com/download/ (#52002)."));
			return false;
		}
		return GetFunction(hModule, "_GetVers@4", &fpGetVers)
			&& GetFunction(hModule, "_FindUSB@4", &fpFindUSB)
			&& GetFunction(hModule, "_GetTempHumid@12", &fpGetTempHumid)
			&& GetFunction(hModule, "_ControlIO@12", &fpControlIO)
			&& GetFunction(hModule, "_SetHeater@8", &fpSetHeater)
			&& GetFunction(hModule, "_GetTempHumidTrue@12", &fpGetTempHumidTrue);
	}
};

/**
 * BSTR܂BSysFreeString(p)s܂Bshared_ptrp̍폜qƂĎg܂B
 */
void ReleaseBSTR(BSTR p) { ::SysFreeString(p); }


/**
 * ڑĂusbrhfoCX̃foCX񋓂܂B
 */
void EnumerateDevices(std::vector<boost::shared_ptr<OLECHAR> > *devices)
{
	LONG index = 0;
	for(;;){
		BSTR dev = usbmeter::FindUSB(&index); //UnicoderhłANSIŕԂĂ̂ŒӁB
		// CSimpleArray<CAdapt<CComBSTR> >Ȃ񂩂gׂȂ̂ȂATLɂ͏ڂȂRs[ƂCɂȂ̂ŁAƂ肠std::vectorshared_ptrgƂɂB
		// std::vector<BSTR>ŉ͍sȂĂ̂肩BǂfoCX񋓂͈̂񂾂B
		boost::shared_ptr<OLECHAR> devptr(dev, ReleaseBSTR);
		if(::SysStringLen(dev) == 0){
			break;
		}
		devices->push_back(devptr);
	}
	return;
}









// --------------------------------------------------------------------------
// C_CAO{bNX
// --------------------------------------------------------------------------

/**
 * AvP[V̒SƂȂ_CAO{bNXEBhEłB
 *
 * _CAOɂ݂͌̌vf[^ݒ\܂B
 * ^C}[CxgŌv⃍Oo͂s܂B
 * N̓^XNACRo^Ĕ\ԂɂȂ܂B
 */
class MainDialogBox
{
	static const UINT WM_NOTIFY_ICON = (WM_USER + 100);///< VXegCACR̒ʒmpB
	static const UINT TIMER_ID = 1; ///< j^XVAoߎԊmFɎgp^C}[IDB
	static const UINT TIMER_PERIOD = 1000; ///< j^XVpxAoߎԊmFpxB

	const HINSTANCE hInst_;
	const std::vector<boost::shared_ptr<OLECHAR> > &devices_;

	HWND hWnd_;
	HMENU hPopupMenu_;
	NOTIFYICONDATA NotifyIconData_;
	HICON hIconApp_;

	SYSTEMTIME prevTime_;

	bool visible_;

	typedef std::basic_string<TCHAR> StdString;
	StdString outputFilename_;  //example: temp_humid.csv
	StdString outputHTTPURL_;   //example: example.jp/temp_humid/record.pl
public:
	MainDialogBox(HINSTANCE hInst, const std::vector<boost::shared_ptr<OLECHAR> > &devices)
		: hInst_(hInst), devices_(devices), hWnd_(NULL), hPopupMenu_(), hIconApp_(), visible_()
		, outputFilename_()
		, outputHTTPURL_()
	{}
	virtual ~MainDialogBox(){}

	HWND Create(void)
	{
		if(hWnd_){
			return NULL;
		}
		hWnd_ = ::CreateDialogParam(
			hInst_,
			MAKEINTRESOURCE(IDD_MAIN),
			NULL,
			(DLGPROC)DlgProcStatic,
			(LPARAM)this);
		return hWnd_;
	}


private:
	static INT_PTR CALLBACK DlgProcStatic(
		HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
	{
		if(msg == WM_INITDIALOG){
			::SetWindowLong(hWnd, DWL_USER, lParam);
		}
		MainDialogBox *pThis;
		pThis = (MainDialogBox *)::GetWindowLong(hWnd, DWL_USER);

		if(pThis){
			return pThis->DlgProc(hWnd, msg, wParam, lParam);
		}
		else{
			return FALSE;
		}
	}

	/**
	 * _CAObZ[WbZ[We֐ɐU蕪邾łB
	 * @return TRUE(1)̂ƂbZ[W
	 */
	INT_PTR DlgProc(
		HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam)
	{
		switch(Message){
		case WM_INITDIALOG:
			return OnInitDialog(hWnd, (HWND)wParam, lParam);
		case WM_DESTROY:
			return OnDestroy(hWnd);
		case WM_CLOSE:
			return OnClose(hWnd);
		case WM_COMMAND:
			return OnCommand(
				hWnd,
				(int)(LOWORD(wParam)),
				(HWND)(lParam),
				(UINT)HIWORD(wParam));
		case WM_NOTIFY_ICON:
			OnNotifyIcon(hWnd, wParam, lParam);
			return TRUE;
		case WM_TIMER:
			OnTimer(hWnd, wParam);
			return 0;
		}
		return FALSE;
	}

	/**
	 * WM_INITDIALOGbZ[W̏
	 */
	BOOL OnInitDialog(HWND hWnd, HWND hWndFocus, LPARAM lParam)
	{
		// ݒ̓ǂݍ
		LoadSettingsFromRegistory();
		CopySettingsVariablesToDlgItems();

		// |bvAbvj[쐬
		hPopupMenu_ = ::LoadMenu(hInst_, _T("POPUPMENU"));

		// ACRǂݍ
		hIconApp_ = ::LoadIcon(hInst_, _T("IDI_APP"));

		// ACRݒ
		::ZeroMemory(&NotifyIconData_, sizeof(NotifyIconData_));
		NotifyIconData_.cbSize = sizeof(NotifyIconData_);
		NotifyIconData_.hWnd   = hWnd;
		NotifyIconData_.uID    = NULL;
		NotifyIconData_.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
		NotifyIconData_.uCallbackMessage = WM_NOTIFY_ICON;
		NotifyIconData_.hIcon  = hIconApp_;
		::lstrcpy(NotifyIconData_.szTip, _T("USBRH Thermo-Hygrometer"));

		::Shell_NotifyIcon(NIM_ADD, &NotifyIconData_);

		// ^C}[ݒ
		::GetLocalTime(&prevTime_);
		::SetTimer(hWnd, TIMER_ID, TIMER_PERIOD, NULL);

		return TRUE;
	}

	/**
	 * WM_DESTROYbZ[W
	 */
	int OnDestroy(HWND hWnd)
	{
		PostQuitMessage(0);

		// |bvAbvj[폜
		DestroyMenu(hPopupMenu_);

		// CWQ[^̈悩폜
		Shell_NotifyIcon(NIM_DELETE, &NotifyIconData_);

		// ^C}[폜
		KillTimer(hWnd, TIMER_ID);

		return TRUE;
	}

	/**
	 * _CAO{bNX\Ԃɂ܂B
	 */
	void Show(void)
	{
		if(!visible_){
			visible_ = true;
			ShowWindow(hWnd_, SW_SHOW);
		}
	}

	/**
	 * _CAO{bNX\Ԃɂ܂B
	 */
	void Hide(void)
	{
		if(visible_){
			visible_ = false;
			ShowWindow(hWnd_, SW_HIDE);
		}
	}

	/**
	 * WM_CLOSEbZ[W
	 */
	int OnClose(HWND hWnd)
	{
		Hide();
		return TRUE;
	}

	/**
	 * WM_COMMANDbZ[W
	 * @param hWnd EBhEnh	
	 * @param id IDARg[IDAANZ[^ID
	 * @param hWndCtl Rg[̃nh
	 * @param codeNotify ʒmR[h
	 */
	int OnCommand(HWND hWnd,
		int id,
		HWND hwndctl,
		UINT codeNotify)
	{
		switch (id) {
		case IDM_MONITOR:
			Show();
			break;
		case IDM_ABOUT:
			MessageBox(
				hWnd,
				_T("USBRH Monitor\n2010 AKIYAMA Kouhei"),
				_T("^Cg"),
				MB_OK);
			break;
		case IDM_EXIT:
			DestroyWindow(hWnd);
			break;
		// C_CAÕR}h
		case IDOK:
			CopySettingsDlgItemsToVariables();
			StoreSettingsToRegistory();
			OnClose(hWnd);
			return TRUE;
		case IDCANCEL:
			CopySettingsVariablesToDlgItems();
			return TRUE;
		}

		return TRUE;
	}


	/**
	 * VXegCbZ[W
	 * WM_NOTIFY_ICONbZ[W
	 * @param hWnd EBhEnh
	 * @param wParam
	 * @param lParam bZ[W
	 */
	void OnNotifyIcon(HWND hWnd, WPARAM wParam, LPARAM lParam)
	{
		switch(lParam){
		case WM_LBUTTONDOWN:
		case WM_RBUTTONDOWN:
			POINT pt;
			GetCursorPos(&pt);

			HMENU hMenu = GetSubMenu(hPopupMenu_, 0);
			SetForegroundWindow( hWnd ) ;

			TrackPopupMenu(hMenu
				, TPM_LEFTBUTTON | TPM_BOTTOMALIGN | TPM_RIGHTALIGN
				, pt.x, pt.y, 0, hWnd, NULL);

			break;
		}
	}


	// ݒ̕ۑƓǂݍ
	static const TCHAR *GetAppRegistryKey(void)
	{
		return _T("SOFTWARE\\Misohena Laboratories\\usbrh_monitor");
	}

	void LoadSettingsFromRegistory(void)
	{
		struct Reg {
			static StdString ReadString(HKEY hKey, const TCHAR *name, const TCHAR *defaultValue = _T(""))
			{
				const UINT MAX_STRING_LENGTH = 1024;
				TCHAR value[MAX_STRING_LENGTH + 1];
				ZeroMemory(value, sizeof(value));
				DWORD type = 0;
				DWORD bytes = MAX_STRING_LENGTH * sizeof(TCHAR);
				LONG result = RegQueryValueEx(hKey, name, NULL, &type, (BYTE *)value, &bytes);
				if(result == ERROR_SUCCESS && type == REG_SZ){
					return StdString(value);
				}
				return defaultValue;
			}
		};
		HKEY hKey;
		if(::RegOpenKeyEx(HKEY_CURRENT_USER, GetAppRegistryKey(), NULL, KEY_ALL_ACCESS, &hKey) != ERROR_SUCCESS){
			return; //error
		}
		outputFilename_ = Reg::ReadString(hKey, "OutputFilename");
		outputHTTPURL_ = Reg::ReadString(hKey, "OutputHTTPURL");
		RegCloseKey(hKey);
	}

	void StoreSettingsToRegistory(void)
	{
		struct Reg {
			static void WriteString(HKEY hKey, const TCHAR *name, const StdString &value)
			{
				const DWORD bytes = (DWORD)((value.size() + 1) * sizeof(TCHAR));
				LONG result = RegSetValueEx(hKey, name, NULL, REG_SZ, (BYTE *)value.c_str(), bytes);
				//if(result == ERROR_SUCCESS){
			}
		};
		HKEY hKey;
		DWORD disposition;
		if(::RegCreateKeyEx(HKEY_CURRENT_USER, GetAppRegistryKey(), NULL, _T(""), REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, &disposition) != ERROR_SUCCESS){
			return; // error
		}
		Reg::WriteString(hKey, "OutputFilename", outputFilename_);
		Reg::WriteString(hKey, "OutputHTTPURL", outputHTTPURL_);
		RegCloseKey(hKey);
	}

	/**
	 * ݒoϐ_CAOփRs[܂B
	 * WXgǂݍ񂾌LZ̎ȂǂɎgp܂B
	 */
	void CopySettingsVariablesToDlgItems(void)
	{
		SetDlgItemText(hWnd_, IDC_EDIT_OUTPUT_FILE, outputFilename_.c_str());
		SetDlgItemText(hWnd_, IDC_EDIT_OUTPUT_URL, outputHTTPURL_.c_str());
	}

	/**
	 * ݒ_CAO烁oϐփRs[܂B
	 * OK{^ƂɎgp܂B
	 */
	void CopySettingsDlgItemsToVariables(void)
	{
		UINT len;
		TCHAR str[MAX_PATH];
		len = GetDlgItemText(hWnd_, IDC_EDIT_OUTPUT_FILE, str, sizeof(str) - 1);
		outputFilename_ = StdString(str, len);

		len = GetDlgItemText(hWnd_, IDC_EDIT_OUTPUT_URL, str, sizeof(str) - 1);
		outputHTTPURL_ = StdString(str, len);
	}


	// ^C}[ɂvƏo


	void OnTimer(HWND hWnd, UINT TimerId)
	{
		SYSTEMTIME currTime;
		::GetLocalTime(&currTime);
		SYSTEMTIME prevTime = prevTime_;
		prevTime_ = currTime;

		if(currTime.wMinute != prevTime.wMinute){ ///@todo ݂͈ꕪԊuŒɂȂĂ̂ŁAςɂB
			Log(currTime, &MainDialogBox::OutputToFile);
			Log(currTime, &MainDialogBox::OutputToHTTP);
		}

		if(visible_){ // _CAO\ĂƂ
			Log(currTime, &MainDialogBox::OutputToDialog);
		}
	}

	void Log(const SYSTEMTIME &currTime, void (MainDialogBox::*func)(const SYSTEMTIME &, double, double))
	{
		if(devices_.empty()){
			return;
		}

		///@todo ̃foCXɑΉBusbrhĂȂ̂őΉCɂȂȂBǂo͂ɂׂ悭ȂB

		double temp = 0;
		double humid = 0;
		if(usbmeter::GetTempHumidTrue(devices_[0].get(), &temp, &humid) != 0){
			return;
		}

		(this->*func)(currTime, temp, humid);
	}

	/**
	 * vf[^t@C֒ǋLo͂܂B
	 */
	void OutputToFile(const SYSTEMTIME &currTime, double temp, double humid)
	{
		if(outputFilename_.empty()){
			return;
		}
		char str[512];
		const int size = std::sprintf(
			str, "%04d/%02d/%02d %02d:%02d:%02d,%lf,%lf\r\n",
			currTime.wYear,
			currTime.wMonth,
			currTime.wDay,
			currTime.wHour,
			currTime.wMinute,
			currTime.wSecond,
			temp,
			humid);
		HANDLE hFile = ::CreateFile(outputFilename_.c_str(), GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		::SetFilePointer(hFile, 0, NULL, FILE_END);
		DWORD written = 0;
		::WriteFile(hFile, str, size, &written, NULL);
		::CloseHandle(hFile);
	}

	/**
	 * vf[^_CAO{bNX֏o͂܂B
	 */
	void OutputToDialog(const SYSTEMTIME &currTime, double temp, double humid)
	{
		TCHAR str[1024];
		std::sprintf(
			str, "%s\n%04d/%02d/%02d %02d:%02d:%02d\ntemp=%lf humid=%lf",
			devices_[0].get(),
			currTime.wYear,
			currTime.wMonth,
			currTime.wDay,
			currTime.wHour,
			currTime.wMinute,
			currTime.wSecond,
			temp,
			humid);

		SetDlgItemText(hWnd_, IDC_STATIC_STATUS, str);
	}

	/**
	 * vf[^HTTPGET\bh̃p[^Ƃďo͂܂B
	 * http://example.jp/record.cgi?temp=25.543&humid=45.123 ̂悤Ȍ`ŃANZX܂B
	 */
	void OutputToHTTP(const SYSTEMTIME &currTime, double temp, double humid)
	{
		if(outputHTTPURL_.empty()){
			return;
		}

		///@todo |[gwł悤ɂB
		///@todo 擪 http:// ƓĂvȂ悤ɂB
		const StdString::size_type firstSlashPos = outputHTTPURL_.find_first_of('/');
		const StdString destHostName = (firstSlashPos != StdString::npos) ? outputHTTPURL_.substr(0, firstSlashPos) : outputHTTPURL_;
		const StdString destURL      = (firstSlashPos != StdString::npos) ? outputHTTPURL_.substr(firstSlashPos) : "/";

		struct Local
		{
			static bool connectToHost(SOCKET &sock, const char *hostname)
			{
				sockaddr_in server;
				server.sin_family = AF_INET;
				server.sin_port = htons(80);
				server.sin_addr.S_un.S_addr = inet_addr(hostname);

				if(server.sin_addr.S_un.S_addr == 0xffffffff) {
					hostent *host = gethostbyname(hostname);
					if(host == NULL){
						return false;
					}

					unsigned int **addrptr = (unsigned int **)host->h_addr_list;
					for(; *addrptr != NULL; ++addrptr){
						server.sin_addr.S_un.S_addr = *(*addrptr);

						if(connect(sock, (sockaddr *)&server, sizeof(server)) == 0){
							break;
						}
					}
					if(*addrptr == NULL){
						return false;
					}
				}
				else{
					if (connect(sock, (sockaddr *)&server, sizeof(server)) != 0){
						return false;
					}
				}
				return true;
			}
		};

		SOCKET sock;
		sock = socket(AF_INET, SOCK_STREAM, 0);
		if(sock != INVALID_SOCKET) {

			// ڑ
			if(Local::connectToHost(sock, destHostName.c_str())){

				// NGXg쐬
				char requestParams[1024];
				std::sprintf(requestParams, "?temp=%lf&humid=%lf", temp, humid);
				std::string request = std::string("GET ") + destURL + requestParams + " HTTP/1.0\r\n\r\n";

				// HTTPNGXgM
				bool error = false;
				int n = send(sock, request.c_str(), request.size(), 0);
				if(n < 0) {
					error = true;
				}

				// HTTPbZ[WM
				std::string result;
				while(n > 0){
					char buf[128];
					n = recv(sock, buf, sizeof(buf), 0);
					if(n < 0) {
						error = true;
						break;
					}

					// M
					result.append(buf, n);
				}
			}

			closesocket(sock);
		}
	}

};// class MainDialogBox


int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpszCmdParam, int nCmdShow)
{
	// xvExvfoCX̏
	if(!usbmeter::LoadUSBMeterDLL()){
		return -1;
	}
	std::vector<boost::shared_ptr<OLECHAR> > devices;
	EnumerateDevices(&devices);
	if(devices.empty()){
		MessageBox(NULL, _T("No USB meter device found."), _T("Error"), MB_OK);
		return -1;
	}

	// WinSock̏
	WSADATA wsaData;
	if(WSAStartup(MAKEWORD(2,0), &wsaData) != 0){
		MessageBox(NULL, _T("WSAStartup failed."), _T("Error"), MB_OK);
		return -1;
	}

	// C_CAO쐬
	MainDialogBox dlg(hInst, devices);
	HWND hWnd = dlg.Create();
	if(!hWnd){
		return -1;
	}
	
	// bZ[W[v
	MSG msg;
	while(GetMessage(&msg, NULL, 0, 0) != 0){
		if(!IsDialogMessage(hWnd, &msg)){
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	// WinSock̉
	WSACleanup();

	return (int)msg.wParam;
}
